/*
# Solaris driver for the following USB ethernet adapters:
# Linksys USB200M, Hawking UF200, TRENDnet TU2-ET100, NetGear FA120, D-Link DUB-E100

# Copyright (C) 2003, 2004 Maxim A. Sadovski (maxim_sadovski at mail dot ru)

# This driver is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This driver is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this driver; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/*
 * Required USB DDK defines
 */

#if defined(lint) && !defined(DEBUG)
#define	DEBUG
#endif

#define	USBDRV_MAJOR_VER	0
#define	USBDRV_MINOR_VER	8


/*
 * Includes, Declarations and Local Data
 */

#include	<sys/varargs.h>
#include        <sys/usb/usba.h>
#include        <sys/miiregs.h>
#include        <sys/ethernet.h>
#include	<sys/types.h>
#include	<sys/errno.h>
#include	<sys/debug.h>
#include	<sys/stropts.h>
#include	<sys/stream.h>
#include	<sys/strlog.h>
#include	<sys/cmn_err.h>
#include	<sys/kmem.h>
#include	<sys/conf.h>
#include	<sys/ksynch.h>
#include	<sys/stat.h>
#include	<sys/dlpi.h>
#include        <sys/gld.h>
#include 	<sys/modctl.h>
#include	<sys/ddi.h>
#include	<sys/sunddi.h>
#include	"usbnet.h"


/*
 * Function prototypes.
 */

static	int             usbnetattach(dev_info_t *, ddi_attach_cmd_t);
static	int             usbnetdetach(dev_info_t *, ddi_detach_cmd_t);

static  int             usbnet_reset(gld_mac_info_t *);
static  int             usbnet_start(gld_mac_info_t *);
static  int             usbnet_stop(gld_mac_info_t *);
static  int             usbnet_set_mac_address(gld_mac_info_t *, uint8_t *);
static  int             usbnet_set_multicast(gld_mac_info_t *, uint8_t *, int);
static  int             usbnet_set_promiscuous(gld_mac_info_t *, int);
static  int             usbnet_get_stats(gld_mac_info_t *, struct gld_stats *);
static  int             usbnet_send(gld_mac_info_t *, mblk_t *);

static void             usbnet_control_pipe_callback(usb_pipe_handle_t pipe_handle, usb_ctrl_req_t *ctrl_request);
static void             usbnet_control_pipe_exception_callback(usb_pipe_handle_t pipe_handle, usb_ctrl_req_t *ctrl_request);

static void             usbnet_interrupt_pipe_callback(usb_pipe_handle_t pipe_handle, usb_intr_req_t *intr_request);
static void             usbnet_interrupt_pipe_exception_callback(usb_pipe_handle_t pipe_handle, usb_intr_req_t *intr_request);

static void             usbnet_bulk_out_pipe_callback(usb_pipe_handle_t pipe_handle, usb_bulk_req_t *bulk_request);
static void             usbnet_bulk_out_pipe_exception_callback(usb_pipe_handle_t pipe_handle, usb_bulk_req_t *bulk_request);

static void             usbnet_bulk_in_pipe_callback(usb_pipe_handle_t pipe_handle, usb_bulk_req_t *bulk_request);
static void             usbnet_bulk_in_pipe_exception_callback(usb_pipe_handle_t pipe_handle, usb_bulk_req_t *bulk_request);


/*
 * Make sure the kernel knows the modules we depend upon.
 */

char                    _depends_on[] = {
        "misc/usba10 misc/gld"
};


/*
 * Error message control:
 * 0 = Error messages.
 * 1 = Debug messages.
 * 2 = More debug messages.
 * 3 = A lot more debug messages.
 * Can set the value of the debug flag using adb or in /etc/system file
 * with "set usbnet:usbnetdebug=<value>
 */

static int                      usbnetdebug = 0;


/*
 * Broadcast medium address definition.
 */

static struct medium_addr       broadcastaddr = {
        { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff }
};


/*
 * Physical medium address definition.
 */

static struct medium_addr       physicaladdr = {
        { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff }
};


/*
 * Linked list of "usbnet" structures - one per device.
 */

static struct usbnet    *usbnetdev = NULL;


/* Opaque handle for state structs */

static void             *usbnet_state;


/* Driver information per supported device type */

static struct   {
        uint16_t        usb_vendor_id;
        uint16_t        usb_product_id;
        uint_t          driver_info;
}                       usbnet_driver_info [] = {
        /* Linksys USB200M */
        { 0x077b, 0x2226, 0x00130103 },
        /* Hawking UF200, TRENDnet TU2-ET100 */
        { 0x07b8, 0x420a, 0x001f1d1f },
        /* NetGear FA120 */
        { 0x0846, 0x1040, 0x00130103 },
        /* D-Link DUB-E100 */
        { 0x2001, 0x1a00, 0x009f9d9f },
        /* End of the list */
        { 0x0, 0x0, 0x0 }
};


/* Standard Streams declarations */

static	struct	module_info	usbnetminfo = {
	USBNETIDNUM,	/* mi_idnum */
	USBNETNAME,	/* mi_idname */
	USBNETMINPSZ,	/* mi_minpsz */
	USBNETMAXPSZ,	/* mi_maxpsz */
	USBNETHIWAT,	/* mi_hiwat */
	USBNETLOWAT,	/* mi_lowat */
};

static	struct	qinit	usbnetrinit = {
	0,		/* qi_putp */
	gld_rsrv,       /* qi_srvp */
	gld_open,       /* qi_qopen */
	gld_close,      /* qi_qclose */
	0,		/* qi_qadmin */
	&usbnetminfo,	/* qi_minfo */
	NULL,		/* qi_mstat */
};

static	struct	qinit	usbnetwinit = {
	gld_wput,       /* qi_putp */
	gld_wsrv,       /* qi_srvp */
	0,		/* qi_qopen */
	0,		/* qi_qclose */
	0,		/* qi_qadmin */
	&usbnetminfo,	/* qi_minfo */
	NULL,		/* qi_mstat */
};

static  struct	streamtab	usbnet_info = {
	&usbnetrinit,	/* st_rdinit */
	&usbnetwinit,	/* st_wrinit */
	NULL,		/* st_muxrinit */
	NULL,		/* st_muxwrinit */
};


/* Module Loading/Unloading and Autoconfiguration declarations */

/*
 * cb_ops contains the driver entry points and is roughly equivalent
 * to the cdevsw and bdevsw  structures in previous releases.
 *
 * dev_ops contains, in addition to the pointer to cb_ops, the routines
 * that support loading and unloading our driver.
 *
 * Unsupported entry points are set to nodev, except for the poll
 * routine , which is set to nochpoll(), a routine that returns ENXIO.
 */

static  struct  cb_ops          usbnet_cb_ops = {
	nodev,			/* cb_open */
	nodev,			/* cb_close */
	nodev,			/* cb_strategy */
	nodev,			/* cb_print */
	nodev,			/* cb_dump */
	nodev,			/* cb_read */
	nodev,			/* cb_write */
	nodev,			/* cb_ioctl */
	nodev,			/* cb_devmap */
	nodev,			/* cb_mmap */
	nodev,			/* cb_segmap */
	nochpoll,		/* cb_chpoll */
	ddi_prop_op,		/* cb_prop_op */
	&usbnet_info,		/* cb_stream */
	D_MP | D_NEW,		/* cb_flag */
};

static  struct  dev_ops         usbnet_ops = {
	DEVO_REV,		/* devo_rev */
	0,			/* devo_refcnt */
	gld_getinfo,		/* devo_getinfo */
	nulldev,		/* devo_identify */
	nulldev,		/* devo_probe */
	usbnetattach,		/* devo_attach */
	usbnetdetach,		/* devo_detach */
	nodev,			/* devo_reset */
	&usbnet_cb_ops,		/* devo_cb_ops */
	(struct bus_ops *)NULL,	/* devo_bus_ops */
};


/*
 * Module linkage information for the kernel.
 */

static  struct  modldrv         modldrv = {
	&mod_driverops,                                 /* Type of module.  This one is a driver */
	"USBNet Linksys USB200M Driver 1.0",            /* Description */
	&usbnet_ops,                                    /* Driver options */
};

static  struct  modlinkage      modlinkage = {
	MODREV_1,
        { &modldrv, NULL }
};


/*
 * Print Debugging Output.
 */

static void
usbneterror(const char *fmt, ...)
{
        va_list         args;

        (void) printf("%s: ", USBNETNAME);

        va_start(args, fmt);
        (void) vprintf(fmt, args);
        va_end(args);
}


/*
 * Module Loading and Installation Routines .
 */

/*
 * Module Installation
 * Install the driver and "pre-allocate" space for instances of the driver
 */

int
_init(void)
{
	int	status = 0;

        if (usbnetdebug > 0)
                usbneterror("_init STARTED\n", status);

	if ((status = ddi_soft_state_init(&usbnet_state,
			sizeof (struct usbnet), 0)) != 0) {
                usbneterror("_init - ddi_soft_state_init failed: 0x%x\n", status);
		return (status);
	}

        if ((status = mod_install(&modlinkage)) != 0) {
                usbneterror("_init - mod_install failed: 0x%x\n", status);
                ddi_soft_state_fini(&usbnet_state);
                return (status);
        }

	return (status);
}


/*
 * Module Removal
 */

int
_fini(void)
{
	int	status = 0;

        if (usbnetdebug > 0)
                usbneterror("_fini STARTED\n", status);

	/*
	 *  At this point, make sure you have canceled all call backs and
	 *  de-allocated all memory
	 */
	if ((status = mod_remove(&modlinkage)) != 0) {
                usbneterror("_fini- mod_remove failed: 0x%x\n", status);
		return (status);
	}
	ddi_soft_state_fini(&usbnet_state);

	return (status);
}


/*
 * Return Module Info.
 */

int
_info(struct modinfo *modinfop)
{
        if (usbnetdebug > 0)
                usbneterror("_info STARTED\n");

	return (mod_info(&modlinkage, modinfop));
}


/*
 * ------------------------------------------------------
 *	Capture data from the message, traversing
 *              all the message blocks if necessary
 */

static int
usbnet_glue_stream_msg(mblk_t *mblkptr,                 /* One message block of the message */
                        uint_t size,                    /* Maximum size of the captured data */
                        void *data, uint_t *len)        /* Buffer to store the data and its actual length */
{
        if (mblkptr == NULL)
                return (DDI_FAILURE);

        *len = 0;

        while (mblkptr != NULL && *len < size) {
                int     olen = *len;
                int     blen = mblkptr->b_wptr - mblkptr->b_rptr;

                if (*len + blen > size)
                        blen = size - *len;

                *len += blen;

                bcopy((uint8_t *) mblkptr->b_rptr,
                        (uint8_t *) data + olen,
                        blen);

                mblkptr = mblkptr->b_cont;
        }

        return (DDI_SUCCESS);
}


/*
 * ------------------------------------------------------
 *	Control Pipe Communication Routines
 */

static int
usbnet_ctrl_read(struct usbnet *usbnetp, uint8_t request,
                        uint16_t value, uint16_t index,
                        void *data, uint_t len)
{
        int                     rc;
        uint_t                  rlen;
        mblk_t                  *mblkptr;
        usb_cr_t                comp_rsn;
        usb_cb_flags_t          cb_flags;
        usb_ctrl_setup_t        ctrl_setup;

        if (usbnetdebug > 0)
                usbneterror("usbnet_ctrl_read STARTED. Request: %d Value: %d Index: %d Len: %d\n",
                                request, value, index, len);

        ctrl_setup.bmRequestType = USB_DEV_REQ_DEV_TO_HOST |
                                        USB_DEV_REQ_TYPE_VENDOR |
                                        USB_DEV_REQ_RCPT_DEV;
        ctrl_setup.bRequest = request;
        ctrl_setup.wValue = value;
        ctrl_setup.wIndex = index;
        ctrl_setup.wLength = len;
        ctrl_setup.attrs = 0;

        rc = usb_pipe_ctrl_xfer_wait(usbnetp->usb_control_pipe,
                                        &ctrl_setup,
                                        &mblkptr,
                                        &comp_rsn,
                                        &cb_flags,
                                        0);

        if (rc != USB_SUCCESS) {
                usbneterror("usbnet_ctrl_read: usb_pipe_ctrl_xfer_wait failed: %d. Reason: %d Flag: %d\n",
                                rc, comp_rsn, cb_flags);
                return USB_FAILURE;
        }

        rc = usbnet_glue_stream_msg(mblkptr,
                                        len,
                                        data,
                                        &rlen);

        if (mblkptr != NULL)
                freemsg(mblkptr);

        if (rc != DDI_SUCCESS) {
                usbneterror("usbnet_ctrl_read: bogus message: Mblk: 0x%08p Requested: %d Actual: %d\n",
                                (void *) mblkptr, len, rlen);
                return USB_FAILURE;
        }

        if (rlen < len) {
                usbneterror("usbnet_ctrl_read: short data read: Requested: %d Actual: %d\n",
                                len, rlen);
        }

        return rlen;
}

static int
usbnet_ctrl_read_async(struct usbnet *usbnetp, uint8_t request,
                        uint16_t value, uint16_t index, uint_t len)
{
        int                     rc;
        usb_ctrl_req_t          *control_request;

        if (usbnetdebug > 0)
                usbneterror("usbnet_ctrl_read_async STARTED. Request: %d Value: %d Index: %d Len: %d\n",
                                request, value, index, len);

        control_request = usb_alloc_ctrl_req(usbnetp->usbnet_dip,
                                                len, USB_FLAGS_SLEEP);

        if (control_request == NULL) {
                usbneterror("usbnet_ctrl_read_async: usb_alloc_ctrl_req failed: %d\n",
                                0);
                return USB_FAILURE;
        }

        control_request->ctrl_bmRequestType = USB_DEV_REQ_DEV_TO_HOST |
                                                USB_DEV_REQ_TYPE_VENDOR |
                                                USB_DEV_REQ_RCPT_DEV;
        control_request->ctrl_bRequest = request;
        control_request->ctrl_wValue = value;
        control_request->ctrl_wIndex = index;
        control_request->ctrl_wLength = len;
        control_request->ctrl_timeout = USBNET_DEFAULT_TIMEOUT;
        control_request->ctrl_client_private = (usb_opaque_t) usbnetp;
        control_request->ctrl_cb = usbnet_control_pipe_callback;
        control_request->ctrl_exc_cb = usbnet_control_pipe_exception_callback;

        rc = usb_pipe_ctrl_xfer(usbnetp->usb_control_pipe,
                                        control_request,
                                        0);

        if (rc != USB_SUCCESS) {
                usbneterror("usbnet_ctrl_read_async: usb_pipe_ctrl_xfer failed: %d\n",
                                rc);
                return USB_FAILURE;
        }

        return USB_SUCCESS;
}

static int
usbnet_ctrl_write(struct usbnet *usbnetp, uint8_t request,
                        uint16_t value, uint16_t index,
                        void *data, uint_t len)
{
        int                     rc;
        mblk_t                  *mblkptr;
        usb_cr_t                comp_rsn;
        usb_cb_flags_t          cb_flags;
        usb_ctrl_setup_t        ctrl_setup;

        if (usbnetdebug > 0)
                usbneterror("usbnet_ctrl_write STARTED. Request: %d Value: %d Index: %d Len: %d\n",
                                request, value, index, len);

        mblkptr = allocb(len, 0);

        if (mblkptr == NULL) {
                usbneterror("usbnet_ctrl_write: allocb falied: %d\n",
                                len);
                return USB_FAILURE;
        }

        if (len > 0) {
                bcopy((uint8_t *) data,
                        (uint8_t *) mblkptr->b_wptr,
                        len);

                mblkptr->b_wptr += len;
        }

        ctrl_setup.bmRequestType = USB_DEV_REQ_HOST_TO_DEV |
                                        USB_DEV_REQ_TYPE_VENDOR |
                                        USB_DEV_REQ_RCPT_DEV;
        ctrl_setup.bRequest = request;
        ctrl_setup.wValue = value;
        ctrl_setup.wIndex = index;
        ctrl_setup.wLength = len;
        ctrl_setup.attrs = 0;

        rc = usb_pipe_ctrl_xfer_wait(usbnetp->usb_control_pipe,
                                        &ctrl_setup,
                                        &mblkptr,
                                        &comp_rsn,
                                        &cb_flags,
                                        0);

        if (rc != USB_SUCCESS) {
                usbneterror("usbnet_ctrl_write: usb_pipe_ctrl_xfer_wait failed: %d. Reason: %d Flag: %d\n",
                                rc, comp_rsn, cb_flags);
                return USB_FAILURE;
        }

        return len;
}

static int
usbnet_ctrl_write_async(struct usbnet *usbnetp, uint8_t request,
                        uint16_t value, uint16_t index,
                        void *data, uint_t len)
{
        int                     rc;
        usb_ctrl_req_t          *control_request;

        if (usbnetdebug > 0)
                usbneterror("usbnet_ctrl_write_async STARTED. Request: %d Value: %d Index: %d Len: %d\n",
                                request, value, index, len);

        control_request = usb_alloc_ctrl_req(usbnetp->usbnet_dip,
                                                len, USB_FLAGS_SLEEP);

        if (control_request == NULL) {
                usbneterror("usbnet_ctrl_write_async: usb_alloc_ctrl_req failed: %d\n",
                                len);
                return USB_FAILURE;
        }

        control_request->ctrl_bmRequestType = USB_DEV_REQ_HOST_TO_DEV |
                                                USB_DEV_REQ_TYPE_VENDOR |
                                                USB_DEV_REQ_RCPT_DEV;
        control_request->ctrl_bRequest = request;
        control_request->ctrl_wValue = value;
        control_request->ctrl_wIndex = index;
        control_request->ctrl_wLength = len;
        control_request->ctrl_timeout = USBNET_DEFAULT_TIMEOUT;
        control_request->ctrl_client_private = (usb_opaque_t) usbnetp;
        control_request->ctrl_cb = usbnet_control_pipe_callback;
        control_request->ctrl_exc_cb = usbnet_control_pipe_exception_callback;

        if (len > 0) {
                bcopy((uint8_t *) data,
                        (uint8_t *) control_request->ctrl_data->b_wptr,
                        len);

                control_request->ctrl_data->b_wptr += len;
        }

        rc = usb_pipe_ctrl_xfer(usbnetp->usb_control_pipe,
                                        control_request,
                                        0);

        if (rc != USB_SUCCESS) {
                usbneterror("usbnet_ctrl_write_async: usb_pipe_ctrl_xfer failed: %d",
                                rc);
                return USB_FAILURE;
        }

        return USB_SUCCESS;
}


/*
 * ------------------------------------------------------
 *	USB Pipe Callbacks
 */

static void
usbnet_dump_data(void *ptr, int len)
{
        int             count = 0;

        for (count = 0; len > 0; count ++)
        {
                int             ind;
                int             minlen;
                char            buf[3 * USBNET_HEX_BYTES_PER_LINE + 4];

                if (len < USBNET_HEX_BYTES_PER_LINE)
                      minlen = len; else
                      minlen = USBNET_HEX_BYTES_PER_LINE;

                len -= minlen;

                for (ind = 0; ind < minlen; ind ++)
                        sprintf(&buf[3 * ind], "%02X ", *((uchar_t *) ptr + USBNET_HEX_BYTES_PER_LINE * count + ind));

                usbneterror("%s\n", buf);
        }
}

static void
usbnet_control_pipe_exception_callback(usb_pipe_handle_t pipe_handle, usb_ctrl_req_t *ctrl_request)
{
        mblk_t                  *mblkptr = ctrl_request->ctrl_data;
        struct usbnet           *usbnetp = (struct usbnet *) ctrl_request->ctrl_client_private;

        if (usbnetdebug > 0)
                usbneterror("usbnet_control_pipe_exception_callback STARTED. Pipe Handle: 0x%08p Usbnetp: 0x%08p Mblk: 0x%08p Reason: %d Flag: %d\n",
                                (void *) pipe_handle,
                                (void *) usbnetp,
                                (void *) mblkptr,
                                ctrl_request->ctrl_completion_reason,
                                ctrl_request->ctrl_exc_cb_flags);

        usb_free_ctrl_req(ctrl_request);

	mutex_enter(&usbnetp->usbnet_intrlock);
        usbnetp->usb_link_state = USBNET_PHY_STATE_NO_LINK;
        mutex_exit(&usbnetp->usbnet_intrlock);

        return;
}

static void
usbnet_control_pipe_callback(usb_pipe_handle_t pipe_handle, usb_ctrl_req_t *ctrl_request)
{
        int                     rc;
        uint_t                  len;
        uint8_t                 buf8[16];
        uint16_t                mii_status = 0;
        uint16_t                full_duplex = 0;
        uint16_t                flow_control = 0;
        uint16_t                fast_speed = 0;
        mblk_t                  *mblkptr = ctrl_request->ctrl_data;
        struct usbnet           *usbnetp = (struct usbnet *) ctrl_request->ctrl_client_private;

        if (usbnetdebug > 0)
                usbneterror("usbnet_control_pipe_callback STARTED. Pipe Handle: 0x%08p Usbnetp: 0x%08p Mblk: 0x%08p Reason: %d Flag: %d\n",
                                (void *) pipe_handle,
                                (void *) usbnetp,
                                (void *) mblkptr,
                                ctrl_request->ctrl_completion_reason,
                                ctrl_request->ctrl_exc_cb_flags);

        rc = usbnet_glue_stream_msg(mblkptr,
                                        sizeof(buf8),
                                        buf8,
                                        &len);

        /* No need to have this any more... */
        usb_free_ctrl_req(ctrl_request);

	mutex_enter(&usbnetp->usbnet_intrlock);

        /* Streams currently being closed */
        if ((usbnetp->usbnet_flags & USBNET_RUNNING) == 0) {
		usbnetp->usb_link_state = USBNET_PHY_STATE_NO_LINK;
                mutex_exit(&usbnetp->usbnet_intrlock);
                return;
	}

	switch (usbnetp->usb_link_state) {

                /* Now that we're in software MII mode, read the BMSR */
                case USBNET_PHY_STATE_POLLING_1:

                        usbnetp->usb_link_state = USBNET_PHY_STATE_POLLING_2;

                        mutex_exit(&usbnetp->usbnet_intrlock);

                        usbnet_ctrl_read_async(usbnetp,
                                                USBNET_CMD_READ_MII_REG,
                                                usbnetp->usb_physical_id[1],
                                                MII_STATUS, 2);

                        return;

		/* Done reading BMSR */
                case USBNET_PHY_STATE_POLLING_2:

                        if (rc == DDI_SUCCESS && len >= sizeof(uint16_t)) {

                                mii_status = *((uint16_t *) buf8);

                                if ((mii_status & (MII_STATUS_LINKUP | MII_STATUS_CANAUTONEG)) ==
                                        (MII_STATUS_LINKUP | MII_STATUS_CANAUTONEG)) {

                                        if (mii_status & MII_STATUS_ANDONE) {

                                                /* Autonegotiation done, go on to read LPA */
                                                usbnetp->usb_link_state = USBNET_PHY_STATE_POLLING_3;

                                                mutex_exit(&usbnetp->usbnet_intrlock);

                                                usbnet_ctrl_read_async(usbnetp,
                                                                        USBNET_CMD_READ_MII_REG,
                                                                        usbnetp->usb_physical_id[1],
                                                                        MII_AN_LPABLE, 2);

                                                return;

                                        }

			        }

                        }
			/* else fall through to abort */

        		usbnetp->usb_link_state = USBNET_PHY_STATE_ABORT_POLL;

                /* Got what we needed from PHY, set back to hardware MII mode
                   (Do same for abort in mid-poll) */
        	case USBNET_PHY_STATE_POLLING_3:

                        /* There should be LPA data by now... */
                        if (rc == DDI_SUCCESS && len >= sizeof(uint16_t) &&
                                usbnetp->usb_link_state == USBNET_PHY_STATE_POLLING_3) {

                                usbnetp->usb_temp = *((uint16_t *) buf8);

                        } else {

                                usbnetp->usb_link_state = USBNET_PHY_STATE_ABORT_POLL;

                        }
                        /* fall through */

	        case USBNET_PHY_STATE_ABORT_POLL:

        		usbnetp->usb_link_state ++;

                        mutex_exit(&usbnetp->usbnet_intrlock);

                        usbnet_ctrl_write_async(usbnetp,
                                                USBNET_CMD_SET_HW_MII,
                                                0, 0,
                                                buf8, 0);

                        return;

                /* All done */
                case USBNET_PHY_STATE_SETTING_MAC:

		        usbnetp->usb_link_state = USBNET_PHY_STATE_LINK;

                        mutex_exit(&usbnetp->usbnet_intrlock);

                        return;

                /* The end result, set the right duplex and flow control mode in the
                   MAC (based on the PHY's LPA reg, which should still be in the data
                   buffer) */
        	case USBNET_PHY_STATE_POLLING_4:

                        /* Restore saved LPA data... */
                        mii_status = usbnetp->usb_temp;

                        usbnetp->usb_link_state = USBNET_PHY_STATE_SETTING_MAC;

                        full_duplex = mii_status & (MII_ABILITY_10BASE_T_FD |
                                                        MII_ABILITY_100BASE_TX_FD);

                        fast_speed = mii_status & (MII_ABILITY_100BASE_TX |
                                                        MII_ABILITY_100BASE_TX_FD |
                                                        MII_ABILITY_100BASE_T4);

                        /* Apparently, 0x0400 - Flow Control */
                        flow_control = full_duplex && (mii_status & 0x0400);

                        usbneterror("Link Established: Speed: %d Duplex: %s Flow Control: %s\n",
                                        (fast_speed ? 100 : 10),
                                        (full_duplex ? "Full" : "Half"),
                                        (flow_control ? "On" : "Off"));

                        mutex_exit(&usbnetp->usbnet_intrlock);

                        usbnet_ctrl_write_async(usbnetp,
                                                USBNET_CMD_WRITE_MEDIUM_MODE,
                                                0x04 |
                                                (full_duplex ? 0x02 : 0) |
                                                (flow_control ? 0x10 : 0),
                                                0,
                                                buf8, 0);

                        return;

        	default:

                        usbneterror("Illegal Link State: %d Usbnet: 0x%08p Mblk: 0x%08p. Aborting...\n",
                                        usbnetp->usb_link_state,
                                        (void *)usbnetp,
                                        (void *)mblkptr);
                        /* fall through */

        	case USBNET_PHY_STATE_ABORTING:

                        usbnetp->usb_link_state = USBNET_PHY_STATE_NO_LINK;

                        mutex_exit(&usbnetp->usbnet_intrlock);

                        return;
	}

        ASSERT(0);
}

static void
usbnet_interrupt_pipe_exception_callback(usb_pipe_handle_t pipe_handle, usb_intr_req_t *intr_request)
{
        mblk_t                  *mblkptr = intr_request->intr_data;
        struct usbnet           *usbnetp = (struct usbnet *) intr_request->intr_client_private;

        if (usbnetdebug > 0)
                usbneterror("usbnet_interrupt_pipe_exception_callback STARTED. Pipe Handle: 0x%08p Usbnetp: 0x%08p Mblk: 0x%08p Reason: %d Flag: %d\n",
                                (void *) pipe_handle,
                                (void *) usbnetp,
                                (void *) mblkptr,
                                intr_request->intr_completion_reason,
                                intr_request->intr_exc_cb_flags);

        usb_free_intr_req(intr_request);

	mutex_enter(&usbnetp->usbnet_intrlock);

        /* Streams currently being closed */
        if ((usbnetp->usbnet_flags & USBNET_RUNNING) == 0 ||
                usbnetp->usbnet_intr_active == 0) {
		usbnetp->usb_link_state = USBNET_PHY_STATE_NO_LINK;
                usbnetp->usbnet_intr_active = 0;
                cv_signal(&usbnetp->usbnet_cond);
                mutex_exit(&usbnetp->usbnet_intrlock);
                return;
	}

        mutex_exit(&usbnetp->usbnet_intrlock);

        intr_request = usb_alloc_intr_req(usbnetp->usbnet_dip,
                                                0, USB_FLAGS_SLEEP);

        intr_request->intr_len =
                        usbnetp->usb_interrupt_descr.wMaxPacketSize;
        intr_request->intr_client_private = (usb_opaque_t) usbnetp;
        intr_request->intr_attributes = USB_ATTRS_AUTOCLEARING |
                                                USB_ATTRS_SHORT_XFER_OK |
                                                USB_ATTRS_ONE_XFER;
        intr_request->intr_cb = usbnet_interrupt_pipe_callback;
        intr_request->intr_exc_cb = usbnet_interrupt_pipe_exception_callback;

        usb_pipe_intr_xfer(usbnetp->usb_interrupt_pipe,
                                intr_request,
                                0);

        return;
}

static void
usbnet_interrupt_pipe_callback(usb_pipe_handle_t pipe_handle, usb_intr_req_t *intr_request)
{
        int                     rc;
        uint_t                  len;
        uint8_t                 buf8[16];
        uint8_t                 phy_status = 0;
        mblk_t                  *mblkptr = intr_request->intr_data;
        struct usbnet           *usbnetp = (struct usbnet *) intr_request->intr_client_private;

        if (usbnetdebug > 2)
                usbneterror("usbnet_interrupt_pipe_callback STARTED. Pipe Handle: 0x%08p Usbnetp: 0x%08p Mblk: 0x%08p Reason: %d Flag: %d\n",
                                (void *) pipe_handle,
                                (void *) usbnetp,
                                (void *) mblkptr, 
                                intr_request->intr_completion_reason,
                                intr_request->intr_exc_cb_flags);

        rc = usbnet_glue_stream_msg(mblkptr,
                                        sizeof(buf8),
                                        buf8,
                                        &len);

        /* No need to have this any more... */
        usb_free_intr_req(intr_request);

	mutex_enter(&usbnetp->usbnet_intrlock);

        /* Streams currently being closed */
        if ((usbnetp->usbnet_flags & USBNET_RUNNING) == 0 ||
                usbnetp->usbnet_intr_active == 0) {
		usbnetp->usb_link_state = USBNET_PHY_STATE_NO_LINK;
                usbnetp->usbnet_intr_active = 0;
                cv_signal(&usbnetp->usbnet_cond);
                mutex_exit(&usbnetp->usbnet_intrlock);
                return;
	}

	if (usbnetp->usb_link_state == USBNET_PHY_STATE_INITIALIZING) {
		usbnetp->usb_link_state = USBNET_PHY_STATE_NO_LINK;
                mutex_exit(&usbnetp->usbnet_intrlock);
                goto done;
	}

        /* We should be able to read PHY status, at least... */
        if (rc != DDI_SUCCESS || len < 3) {
                mutex_exit(&usbnetp->usbnet_intrlock);
                goto done;
        }

        /* We care only about primary PHY */
        phy_status = buf8[2] & 1;

	if (phy_status == ((usbnetp->usb_link_state == USBNET_PHY_STATE_NO_LINK) ? 0 : 1)) {
                mutex_exit(&usbnetp->usbnet_intrlock);
                goto done;
	}

        usbneterror("Changing Physical Link State: From: %d To: %d\n",
                        ((usbnetp->usb_link_state == USBNET_PHY_STATE_NO_LINK) ? 0 : 1),
                        phy_status);

	if (phy_status == 0) {

		/* Abort is in-progress - poll the PHY if necessary */
                switch (usbnetp->usb_link_state) {
                        case USBNET_PHY_STATE_POLLING_1:
                        case USBNET_PHY_STATE_POLLING_2:
                        case USBNET_PHY_STATE_POLLING_3:
                                usbnetp->usb_link_state = USBNET_PHY_STATE_ABORT_POLL;
                                break;

		        case USBNET_PHY_STATE_POLLING_4:
		        case USBNET_PHY_STATE_SETTING_MAC:
                                usbnetp->usb_link_state = USBNET_PHY_STATE_ABORTING;
                                break;

		        case USBNET_PHY_STATE_LINK:
                                usbnetp->usb_link_state = USBNET_PHY_STATE_NO_LINK;
                                break;

                        default:
			        /* If we're already aborting, continue aborting */
			        break;
                }

        	mutex_exit(&usbnetp->usbnet_intrlock);

        } else {

		/* Note that we only fall into this case if the previous state was
		   USBNET_PHY_STATE_NO_LINK. When the link is reported active while
		   we're still polling, or when we're aborting, the logic above
		   will just return, and we'll check again next int callback. */
		usbnetp->usb_link_state = USBNET_PHY_STATE_POLLING_1;

        	mutex_exit(&usbnetp->usbnet_intrlock);

                usbnet_ctrl_write_async(usbnetp,
                                        USBNET_CMD_SET_SW_MII,
                                        0, 0,
                                        buf8, 0);

        }

done:
        intr_request = usb_alloc_intr_req(usbnetp->usbnet_dip,
                                                0, USB_FLAGS_SLEEP);

        intr_request->intr_len =
                        usbnetp->usb_interrupt_descr.wMaxPacketSize;
        intr_request->intr_client_private = (usb_opaque_t) usbnetp;
        intr_request->intr_attributes = USB_ATTRS_AUTOCLEARING |
                                                USB_ATTRS_SHORT_XFER_OK |
                                                USB_ATTRS_ONE_XFER;
        intr_request->intr_cb = usbnet_interrupt_pipe_callback;
        intr_request->intr_exc_cb = usbnet_interrupt_pipe_exception_callback;

        usb_pipe_intr_xfer(usbnetp->usb_interrupt_pipe,
                                intr_request,
                                0);

        return;
}

static void
usbnet_bulk_out_pipe_exception_callback(usb_pipe_handle_t pipe_handle, usb_bulk_req_t *bulk_request)
{
        mblk_t                  *mblkptr = bulk_request->bulk_data;
        struct usbnet           *usbnetp = (struct usbnet *) bulk_request->bulk_client_private;

        if (usbnetdebug > 0)
                usbneterror("usbnet_bulk_out_pipe_exception_callback STARTED. Pipe Handle: 0x%08p Usbnetp: 0x%08p Mblk: 0x%08p Reason: %d Flag: %d\n",
                                (void *) pipe_handle,
                                (void *) usbnetp,
                                (void *) mblkptr,
                                bulk_request->bulk_completion_reason,
                                bulk_request->bulk_exc_cb_flags);

        usb_free_bulk_req(bulk_request);

        mutex_enter(&usbnetp->usbnet_xmitlock);

        /* Streams currently being closed */
        if ((usbnetp->usbnet_flags & USBNET_RUNNING) == 0 ||
                usbnetp->usbnet_send_active == 0) {
                usbnetp->usbnet_send_active = 0;
                cv_signal(&usbnetp->usbnet_cond);
	}

        usbnetp->usbnet_send_inprogress = 0;

        mutex_exit(&usbnetp->usbnet_xmitlock);

        return;
}

static void
usbnet_bulk_out_pipe_callback(usb_pipe_handle_t pipe_handle, usb_bulk_req_t *bulk_request)
{
        int                     do_send;
        mblk_t                  *mblkptr = bulk_request->bulk_data;
        struct usbnet           *usbnetp = (struct usbnet *) bulk_request->bulk_client_private;

        if (usbnetdebug > 0)
                usbneterror("usbnet_bulk_out_pipe_callback STARTED. Pipe Handle: 0x%08p Usbnetp: 0x%08p Mblk: 0x%08p Reason: %d Flag: %d\n",
                                (void *) pipe_handle,
                                (void *) usbnetp,
                                (void *) mblkptr,
                                bulk_request->bulk_completion_reason,
                                bulk_request->bulk_exc_cb_flags);

        usb_free_bulk_req(bulk_request);

        mutex_enter(&usbnetp->usbnet_xmitlock);

        /* Streams currently being closed */
        if ((usbnetp->usbnet_flags & USBNET_RUNNING) == 0 ||
                usbnetp->usbnet_send_active == 0) {
                usbnetp->usbnet_send_active = 0;
                usbnetp->usbnet_send_inprogress = 0;
                cv_signal(&usbnetp->usbnet_cond);
                mutex_exit(&usbnetp->usbnet_xmitlock);
                return;
	}

        if (usbnetp->usbnet_send_firstptr != NULL) {

                mblkptr = usbnetp->usbnet_send_firstptr;
                usbnetp->usbnet_send_firstptr = mblkptr->b_next;

                if (usbnetp->usbnet_send_firstptr == NULL) {
                        usbnetp->usbnet_send_lastptr = NULL;
                }

                mblkptr->b_next = NULL;

                usbnetp->usbnet_send_inprogress = 1;
                do_send = 1;

        } else {

                usbnetp->usbnet_send_inprogress = 0;
                do_send = 0;

        }

        mutex_exit(&usbnetp->usbnet_xmitlock);

        if (do_send != 0) {

                if (usbnetdebug > 0)
                        usbneterror("usbnet_bulk_out_pipe_callback: SENDING DATA. Pipe Handle: 0x%08p Usbnetp: 0x%08p Mblk: 0x%08p Len: %d\n",
                                (void *) pipe_handle,
                                (void *) usbnetp,
                                (void *) mblkptr,
                                mblkptr->b_wptr - mblkptr->b_rptr);

                bulk_request = usb_alloc_bulk_req(usbnetp->usbnet_dip,
                                                        0, USB_FLAGS_SLEEP);

                bulk_request->bulk_len = mblkptr->b_wptr - mblkptr->b_rptr;
                bulk_request->bulk_data = mblkptr;
                bulk_request->bulk_timeout = USBNET_DEFAULT_TIMEOUT;
                bulk_request->bulk_client_private = (usb_opaque_t) usbnetp;
                bulk_request->bulk_attributes = USB_ATTRS_AUTOCLEARING;
                bulk_request->bulk_cb = usbnet_bulk_out_pipe_callback;
                bulk_request->bulk_exc_cb = usbnet_bulk_out_pipe_exception_callback;

                usb_pipe_bulk_xfer(usbnetp->usb_bulk_out_pipe,
                                        bulk_request,
                                        0);

        } else {

                if (usbnetdebug > 0)
                        usbneterror("usbnet_bulk_out_pipe_callback: NO DATA TO SEND. Pipe Handle: 0x%08p Usbnetp: 0x%08p\n",
                                (void *) pipe_handle,
                                (void *) usbnetp);

        }

        return;
}

static void
usbnet_bulk_in_pipe_exception_callback(usb_pipe_handle_t pipe_handle, usb_bulk_req_t *bulk_request)
{
        mblk_t                  *mblkptr = bulk_request->bulk_data;
        struct usbnet           *usbnetp = (struct usbnet *) bulk_request->bulk_client_private;

        if (usbnetdebug > 1)
                usbneterror("usbnet_bulk_in_pipe_exception_callback STARTED. Pipe Handle: 0x%08p Usbnetp: 0x%08p Mblk: 0x%08p Reason: %d Flag: %d\n",
                                (void *) pipe_handle,
                                (void *) usbnetp,
                                (void *) mblkptr,
                                bulk_request->bulk_completion_reason,
                                bulk_request->bulk_exc_cb_flags);

        usb_free_bulk_req(bulk_request);

	mutex_enter(&usbnetp->usbnet_xmitlock);

        /* Streams currently being closed */
        if ((usbnetp->usbnet_flags & USBNET_RUNNING) == 0 ||
                usbnetp->usbnet_recv_active == 0) {
                usbnetp->usbnet_recv_active = 0;
                cv_signal(&usbnetp->usbnet_cond);
                mutex_exit(&usbnetp->usbnet_xmitlock);
                return;
	}

        mutex_exit(&usbnetp->usbnet_xmitlock);

        bulk_request = usb_alloc_bulk_req(usbnetp->usbnet_dip,
                                                ETHERMAX,
                                                USB_FLAGS_SLEEP);

        bulk_request->bulk_len = ETHERMAX;
        bulk_request->bulk_timeout = USBNET_DEFAULT_TIMEOUT;
        bulk_request->bulk_client_private = (usb_opaque_t) usbnetp;
        bulk_request->bulk_attributes = USB_ATTRS_AUTOCLEARING |
                                                USB_ATTRS_SHORT_XFER_OK;
        bulk_request->bulk_cb = usbnet_bulk_in_pipe_callback;
        bulk_request->bulk_exc_cb = usbnet_bulk_in_pipe_exception_callback;

        usb_pipe_bulk_xfer(usbnetp->usb_bulk_in_pipe,
                                bulk_request,
                                0);

        return;
}

static void
usbnet_bulk_in_pipe_callback(usb_pipe_handle_t pipe_handle, usb_bulk_req_t *bulk_request)
{
        int                     rc;
        uint_t                  len;
        mblk_t                  *mblkptr = bulk_request->bulk_data;
        struct usbnet           *usbnetp = (struct usbnet *) bulk_request->bulk_client_private;
        struct gld_mac_info     *usbnet_mac_info = usbnetp->usbnet_mac_info;

        if (usbnetdebug > 0)
                usbneterror("usbnet_bulk_in_pipe_callback STARTED. Pipe Handle: 0x%08p Usbnetp: 0x%08p Mblk: 0x%08p Reason: %d Flag: %d\n",
                                (void *) pipe_handle,
                                (void *) usbnetp,
                                (void *) mblkptr,
                                bulk_request->bulk_completion_reason,
                                bulk_request->bulk_exc_cb_flags);

        rc = usbnet_glue_stream_msg(mblkptr,
                                        ETHERMAX,
                                        (uint8_t *) usbnetp->usbnet_recv_buf,
                                        &len);

        usb_free_bulk_req(bulk_request);

	mutex_enter(&usbnetp->usbnet_xmitlock);

        /* Streams currently being closed */
        if ((usbnetp->usbnet_flags & USBNET_RUNNING) == 0 ||
                usbnetp->usbnet_recv_active == 0) {
                usbnetp->usbnet_recv_active = 0;
                cv_signal(&usbnetp->usbnet_cond);
                mutex_exit(&usbnetp->usbnet_xmitlock);
                return;
	}

        mutex_exit(&usbnetp->usbnet_xmitlock);

        if (rc != DDI_SUCCESS) {

                usbneterror("usbnet_bulk_in_pipe_callback: bogus message: Pipe Handle: 0x%08p Usbnetp: 0x%08p Mblk: 0x%08p Actual: %d\n",
                                (void *) pipe_handle,
                                (void *) usbnetp,
                                (void *) mblkptr,
                                len);

        } else {

                if (usbnetdebug > 0)
                        usbneterror("usbnet_bulk_in_pipe_callback: RECEIVED DATA. Pipe Handle: 0x%08p Usbnetp: 0x%08p Mblk: 0x%08p Len: %d\n",
                                (void *) pipe_handle,
                                (void *) usbnetp,
                                (void *) mblkptr,
                                len);

                if (usbnetdebug > 2)
                        usbnet_dump_data(usbnetp->usbnet_recv_buf,
                                                len);

                mblkptr = allocb(len, 0);

                bcopy((uint8_t *) usbnetp->usbnet_recv_buf,
                        (uint8_t *) mblkptr->b_wptr,
                        len);

                mblkptr->b_wptr += len;

                gld_recv(usbnet_mac_info, mblkptr);
        }

        bulk_request = usb_alloc_bulk_req(usbnetp->usbnet_dip,
                                                ETHERMAX,
                                                USB_FLAGS_SLEEP);

        bulk_request->bulk_len = ETHERMAX;
        bulk_request->bulk_timeout = USBNET_DEFAULT_TIMEOUT;
        bulk_request->bulk_client_private = (usb_opaque_t) usbnetp;
        bulk_request->bulk_attributes = USB_ATTRS_AUTOCLEARING |
                                                USB_ATTRS_SHORT_XFER_OK;
        bulk_request->bulk_cb = usbnet_bulk_in_pipe_callback;
        bulk_request->bulk_exc_cb = usbnet_bulk_in_pipe_exception_callback;

        usb_pipe_bulk_xfer(usbnetp->usb_bulk_in_pipe,
                                bulk_request,
                                0);

        return;
}

/*
 * ------------------------------------------------------
 *	Autoconfiguration Routines
 */


/*
 *	Close all open USB pipes except the Control Pipe
 *	Assumes that xmit_lock and intr_lock are held on entry
 */

static void
usbnet_close_data_pipes(struct usbnet* usbnetp)
{
        if (usbnetp->usb_bulk_in_pipe != NULL) {

                if (usbnetdebug > 0)
                        usbneterror("usbnet_close_data_pipes: closing BULK IN pipe: Usbnetp: 0x%08p Pipe: 0x%08p\n",
                                        (void *) usbnetp,
                                        (void *) usbnetp->usb_bulk_in_pipe);

                while (usbnetp->usbnet_recv_active != 0)
                        cv_wait(&usbnetp->usbnet_cond,
                                &usbnetp->usbnet_xmitlock);

                usbnetp->usbnet_recv_active = 0;

                usb_pipe_close(usbnetp->usbnet_dip,
                                        usbnetp->usb_bulk_in_pipe,
                                        USB_FLAGS_SLEEP, NULL, NULL);

                usbnetp->usb_bulk_in_pipe = NULL;

        }

        if (usbnetp->usb_bulk_out_pipe != NULL) {

                if (usbnetdebug > 0)
                        usbneterror("usbnet_close_data_pipes: closing BULK OUT pipe: Usbnetp: 0x%08p Pipe: 0x%08p\n",
                                        (void *) usbnetp,
                                        (void *) usbnetp->usb_bulk_out_pipe);

                while (usbnetp->usbnet_send_active != 0 &&
                        usbnetp->usbnet_send_inprogress != 0)
                        cv_wait(&usbnetp->usbnet_cond,
                                &usbnetp->usbnet_xmitlock);

                usbnetp->usbnet_send_active = 0;
                usbnetp->usbnet_send_inprogress = 0;

                usb_pipe_close(usbnetp->usbnet_dip,
                                        usbnetp->usb_bulk_out_pipe,
                                        USB_FLAGS_SLEEP, NULL, NULL);

                usbnetp->usb_bulk_out_pipe = NULL;

        }

        if (usbnetp->usb_interrupt_pipe != NULL) {

                if (usbnetdebug > 0)
                        usbneterror("usbnet_close_data_pipes: closing INTR pipe: Usbnetp: 0x%08p Pipe: 0x%08p\n",
                                        (void *) usbnetp,
                                        (void *) usbnetp->usb_interrupt_pipe);

                while (usbnetp->usbnet_intr_active != 0)
                        cv_wait(&usbnetp->usbnet_cond,
                                &usbnetp->usbnet_intrlock);

                usbnetp->usbnet_intr_active = 0;

                usb_pipe_close(usbnetp->usbnet_dip,
                                        usbnetp->usb_interrupt_pipe,
                                        USB_FLAGS_SLEEP, NULL, NULL);

                usbnetp->usb_interrupt_pipe = NULL;

        }
}


/*
 *	Do complete Soft State cleanup
 */

static void
usbnetcleanup(struct usbnet* usbnetp)
{
        int             instance;
        mblk_t          *mblkptr;

        if (usbnetp == NULL)
                return;

        instance = ddi_get_instance(usbnetp->usbnet_dip);

        if (usbnetdebug > 0)
                usbneterror("usbnetcleanup STARTED. Usbnetp: 0x%08p Instance: %d\n",
                                (void *) usbnetp,
                                instance);

	mutex_enter(&usbnetp->usbnet_intrlock);
	mutex_enter(&usbnetp->usbnet_xmitlock);

        usbnet_close_data_pipes(usbnetp);

        while (usbnetp->usbnet_send_firstptr != NULL) {
                mblkptr = usbnetp->usbnet_send_firstptr;
                usbnetp->usbnet_send_firstptr =
                                usbnetp->usbnet_send_firstptr->b_next;
                mblkptr->b_next = NULL;
                freemsg(mblkptr);
        }

        usbnetp->usbnet_send_firstptr = NULL;
        usbnetp->usbnet_send_lastptr = NULL;

	mutex_exit(&usbnetp->usbnet_xmitlock);
	mutex_exit(&usbnetp->usbnet_intrlock);

	/* unregister with USB */
        if (usbnetp->usbnet_dev_data != NULL) {
                usb_unregister_client(usbnetp->usbnet_dip,
                                        usbnetp->usbnet_dev_data);
                usbnetp->usbnet_dev_data = NULL;
        }

	/* unregister with GLD */
        if (usbnetp->usbnet_mac_info != NULL) {
                gld_unregister(usbnetp->usbnet_mac_info);
                gld_mac_free(usbnetp->usbnet_mac_info);
                usbnetp->usbnet_mac_info = NULL;
        }

        cv_destroy(&usbnetp->usbnet_cond);
	mutex_destroy(&usbnetp->usbnet_xmitlock);
	mutex_destroy(&usbnetp->usbnet_intrlock);

        ddi_soft_state_free(usbnet_state, instance);
}

/*
 * Interface exists: make available by filling in network interface
 * record.  System will initialize the interface when it is ready
 * to accept packets.
 */

static int
usbnetattach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
	int                     rc;
        int                     ind;
        int                     instance;

        uint8_t                 buf8[4];
        uint_t                  gpio_bits;

	struct usbnet           *usbnetp;
        usb_ep_data_t           *ep_data;
        gld_mac_info_t		*usbnet_mac_info;

        if (usbnetdebug > 0)
                usbneterror("attach STARTED. Cmd: %d\n", cmd);

	if (cmd != DDI_ATTACH)
		return (DDI_FAILURE);

	usbnetdebug = ddi_getprop(DDI_DEV_T_ANY, dip, DDI_PROP_DONTPASS,
					"usbnetdebug", usbnetdebug);

	/*
	 * Allocate soft data structure
	 */
	instance = ddi_get_instance(dip);

	if (ddi_soft_state_zalloc(usbnet_state, instance) != DDI_SUCCESS) {
                usbneterror("ddi_soft_state_zalloc: failed\n");
		return (DDI_FAILURE);
	}

	usbnetp = ddi_get_soft_state(usbnet_state, instance);
	usbnetp->usbnet_dip = dip;

	mutex_init(&usbnetp->usbnet_xmitlock, NULL,
                        MUTEX_DRIVER, NULL);

	mutex_init(&usbnetp->usbnet_intrlock, NULL,
                        MUTEX_DRIVER, NULL);

        cv_init(&usbnetp->usbnet_cond, NULL,
                        CV_DRIVER, NULL);

        /*
         * Start USB Hardware Init. Register USB Driver
         */
        rc = usb_register_client(dip, USBDRV_VERSION,
                                        &usbnetp->usbnet_dev_data,
                                        USB_REG_DESCR_IF, 0);

	if (rc != USB_SUCCESS) {
                usbneterror("usb_register_client: failed: %d\n", rc);
                usbnetcleanup(usbnetp);
		return (DDI_FAILURE);
	}

        for (ind = 0, gpio_bits = 0;
                usbnet_driver_info[ind].usb_vendor_id != 0 &&
                        usbnet_driver_info[ind].usb_product_id != 0;
                ind++) {

                if (usbnet_driver_info[ind].usb_vendor_id ==
                                        usbnetp->usbnet_dev_data->dev_descr->idVendor &&
                        usbnet_driver_info[ind].usb_product_id ==
                                        usbnetp->usbnet_dev_data->dev_descr->idProduct) {

                        gpio_bits = usbnet_driver_info[ind].driver_info;
                        break;

                }

        }

        if (gpio_bits == 0) {
                usbneterror("Unsupported Device: Vendor: 0x%x Product: 0x%x\n",
                                usbnetp->usbnet_dev_data->dev_descr->idVendor,
                                usbnetp->usbnet_dev_data->dev_descr->idProduct);
                usbnetcleanup(usbnetp);
		return (DDI_FAILURE);
	}

        usbneterror("Found Supported Device: Vendor: 0x%x Product: 0x%x\n",
                        usbnetp->usbnet_dev_data->dev_descr->idVendor,
                        usbnetp->usbnet_dev_data->dev_descr->idProduct);

        usbnetp->usb_interface = usbnetp->usbnet_dev_data->dev_curr_if;
	usbnetp->usb_control_pipe = usbnetp->usbnet_dev_data->dev_default_ph;

        /*
         * Get INTR Endpoint Descriptor
         */

        ep_data = usb_get_ep_data(dip,
                                        usbnetp->usbnet_dev_data,
                                        usbnetp->usb_interface,
                                        0,                              /* Alternate interface is always 0 */
                                        USB_EP_ATTR_INTR,
                                        USB_EP_DIR_IN);

	if (ep_data == NULL) {
                usbneterror("usb_get_ep_data(INTR): failed\n");
                usb_free_descr_tree(dip, usbnetp->usbnet_dev_data);
                usbnetcleanup(usbnetp);
		return (DDI_FAILURE);
	} else {
                usbnetp->usb_interrupt_descr = ep_data->ep_descr;
        }

        /*
         * Get BULK OUT Endpoint Descriptor
         */

        ep_data = usb_get_ep_data(dip,
                                        usbnetp->usbnet_dev_data,
                                        usbnetp->usb_interface,
                                        0,                              /* Alternate interface is always 0 */
                                        USB_EP_ATTR_BULK,
                                        USB_EP_DIR_OUT);

	if (ep_data == NULL) {
                usbneterror("usb_get_ep_data(BULK OUT): failed\n");
                usb_free_descr_tree(dip, usbnetp->usbnet_dev_data);
                usbnetcleanup(usbnetp);
		return (DDI_FAILURE);
	} else {
                usbnetp->usb_bulk_out_descr = ep_data->ep_descr;
        }

        /*
         * Get BULK IN Endpoint Descriptor
         */

        ep_data = usb_get_ep_data(dip,
                                        usbnetp->usbnet_dev_data,
                                        usbnetp->usb_interface,
                                        0,
                                        USB_EP_ATTR_BULK,
                                        USB_EP_DIR_IN);

	if (ep_data == NULL) {
                usbneterror("usb_get_ep_data(BULK IN): failed\n");
                usb_free_descr_tree(dip, usbnetp->usbnet_dev_data);
                usbnetcleanup(usbnetp);
		return (DDI_FAILURE);
	} else {
                usbnetp->usb_bulk_in_descr = ep_data->ep_descr;
        }

	/* we no longer need the descriptor tree */
	usb_free_descr_tree(dip, usbnetp->usbnet_dev_data);

	/*
	 * Initialize the hardware:
	 * Toggle the GPIOs in a manufacturer/model specific way
	 */
	for (ind = 2; ind >= 0; ind--) {
		rc = usbnet_ctrl_write(usbnetp,
                                        USBNET_CMD_WRITE_GPIOS,
                                        (gpio_bits >> (ind * 8)) & 0xff, 0,
                                        buf8, 0);
		if (rc < 0) {
                        usbneterror("usb_ctrl_write(GPIO): failed: %d\n", rc);
                        usbnetcleanup(usbnetp);
                        return (DDI_FAILURE);
                }

		drv_usecwait(5000);
	}

	rc = usbnet_ctrl_write(usbnetp,
                                USBNET_CMD_WRITE_RX_CTL,
                                0x80, 0, buf8, 0);
	if (rc < 0) {
                usbneterror("usbnet_ctrl_write(WRITE_RX_CTL): failed: %d\n", rc);
                usbnetcleanup(usbnetp);
		return (DDI_FAILURE);
        }

	/* Get the Physical ID */
	rc = usbnet_ctrl_read(usbnetp,
                                USBNET_CMD_READ_PHY_ID,
                                0, 0, buf8, 2);
	if (rc != 2) {
                usbneterror("usbnet_ctrl_write(READ_PHY_ID): failed: %d\n", rc);
                usbnetcleanup(usbnetp);
		return (DDI_FAILURE);
        }

        usbnetp->usb_physical_id[0] = buf8[0];                  /* Secondary Physical ID */
        usbnetp->usb_physical_id[1] = buf8[1];                  /* Primary Physical ID */

        if (usbnetdebug > 0)
                usbneterror("USB Physical ID Primary: %d Secondary: %d\n",
                                usbnetp->usb_physical_id[1],
                                usbnetp->usb_physical_id[0]);

	/*
	 * Figure out our actual MAC address
	 */
        rc = usbnet_ctrl_read(usbnetp,
                                USBNET_CMD_READ_NODE_ID, 0, 0,
                                (uint8_t *) &physicaladdr.medaddr,
                                ETHERADDRL);

	if (rc != ETHERADDRL) {
                usbneterror("usbnet_ctrl_write(READ_NODE_ID): failed: %d\n", rc);
                usbnetcleanup(usbnetp);
		return (DDI_FAILURE);
        }

	/*
	 * Store the local medium address in usbnetp->usbnet_ouraddr.
	 */
        bcopy((uint8_t *) &physicaladdr.medaddr,
                (uint8_t *) &usbnetp->usbnet_ouraddr.medaddr,
                ETHERADDRL);

	/*
	 * Store the broadcast medium address in usbnetp->usbnet_brdcstaddr.
	 */
        bcopy((uint8_t *) &broadcastaddr.medaddr,
                (uint8_t *) &usbnetp->usbnet_brdcstaddr.medaddr,
                ETHERADDRL);

        usbneterror("Hardware MAC Address: %02x-%02x-%02x-%02x-%02x-%02x\n",
                        usbnetp->usbnet_ouraddr.medaddr[0], usbnetp->usbnet_ouraddr.medaddr[1],
                        usbnetp->usbnet_ouraddr.medaddr[2], usbnetp->usbnet_ouraddr.medaddr[3],
                        usbnetp->usbnet_ouraddr.medaddr[4], usbnetp->usbnet_ouraddr.medaddr[5]);

	/*
	 * Stuff private info into dip.
	 */
	ddi_set_driver_private(dip, (caddr_t)usbnetp);

	/*
	 * Link this per-device structure in with the rest.
	 */

	usbnetp->usbnet_nextp = usbnetdev;
	usbnetdev = usbnetp;

	/*
	 * Initialize GDL structure
	 */

        usbnet_mac_info = gld_mac_alloc(dip);

	if (usbnet_mac_info == NULL) {
                usbneterror("gld_mac_alloc failed\n");
                usbnetcleanup(usbnetp);
		return (DDI_FAILURE);
        }

        usbnetp->usbnet_mac_info = usbnet_mac_info;

        usbnet_mac_info->gldm_devinfo              = dip;
        usbnet_mac_info->gldm_private              = (caddr_t) usbnetp;

        usbnet_mac_info->gldm_reset                = usbnet_reset;
        usbnet_mac_info->gldm_start                = usbnet_start;
        usbnet_mac_info->gldm_stop                 = usbnet_stop;
        usbnet_mac_info->gldm_set_mac_addr         = usbnet_set_mac_address;
        usbnet_mac_info->gldm_send                 = usbnet_send;
        usbnet_mac_info->gldm_set_promiscuous      = usbnet_set_promiscuous;
        usbnet_mac_info->gldm_get_stats            = usbnet_get_stats;
        usbnet_mac_info->gldm_ioctl                = NULL;
        usbnet_mac_info->gldm_set_multicast        = usbnet_set_multicast;
        usbnet_mac_info->gldm_intr                 = NULL;
        usbnet_mac_info->gldm_mctl                 = NULL;

        usbnet_mac_info->gldm_ident                = (char *) ddi_driver_name(dip);
        usbnet_mac_info->gldm_type                 = DL_ETHER;
        usbnet_mac_info->gldm_minpkt               = 0;
        usbnet_mac_info->gldm_maxpkt               = ETHERMTU;
        usbnet_mac_info->gldm_addrlen              = ETHERADDRL;
        usbnet_mac_info->gldm_saplen               = -2;
        usbnet_mac_info->gldm_ppa                  = instance;

        usbnet_mac_info->gldm_broadcast_addr       = (uint8_t *) &usbnetp->usbnet_brdcstaddr.medaddr;
        usbnet_mac_info->gldm_vendor_addr          = (uint8_t *) &usbnetp->usbnet_ouraddr.medaddr;

	/*
	 * Register with GLD.
	 */
        rc = gld_register(dip, (char *) ddi_driver_name(dip), usbnet_mac_info);

	if (rc != DDI_SUCCESS) {
                usbneterror("gld_register failed: %d\n", rc);
                usbnetcleanup(usbnetp);
		return (DDI_FAILURE);
        }

	ddi_report_dev(dip);

	return (DDI_SUCCESS);
}

/*
 * Detach - Free resources allocated in attach
 */

/*ARGSUSED*/
static int
usbnetdetach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	struct usbnet	*usbnetp;
	int 		instance;

        if (usbnetdebug > 0)
                usbneterror("detach STARTED. Cmd: %d\n", cmd);

	if (cmd != DDI_DETACH)
		return (DDI_FAILURE);

	instance = ddi_get_instance(dip);

	usbnetp = ddi_get_soft_state(usbnet_state, instance);

	if (usbnetp == NULL)
		return (DDI_FAILURE);

        usbnetcleanup(usbnetp);

	return (DDI_SUCCESS);
}

/*
 * Unsupported GLD callback routines
 */

static  int
usbnet_set_mac_address(gld_mac_info_t *usbnet_mac_info, uint8_t *physicaladdr)
{
	return (GLD_NOTSUPPORTED);
}

/*
 * Supported GLD callback routines
 */

static  int
usbnet_reset(gld_mac_info_t *usbnet_mac_info)
{
	int             rc;
        uint8_t         buf8[16];
        uint16_t        *buf16 = (uint16_t *) buf8;
        struct usbnet   *usbnetp;

        usbnetp = (struct usbnet *) usbnet_mac_info->gldm_private;

	mutex_enter(&usbnetp->usbnet_intrlock);
	mutex_enter(&usbnetp->usbnet_xmitlock);

        usbnetp->usb_link_state = USBNET_PHY_STATE_INITIALIZING;

	rc = usbnet_ctrl_write(usbnetp,
                                USBNET_CMD_SET_SW_MII,
                                0, 0, buf8, 0);
	if (rc < 0) {
                rc = GLD_FAILURE;
		goto done;
	}

	buf16[0] = MII_CONTROL_RESET;
	rc = usbnet_ctrl_write(usbnetp,
                                USBNET_CMD_WRITE_MII_REG,
                                usbnetp->usb_physical_id[1],
                                MII_CONTROL,
                                buf16, 2);
	if (rc < 0) {
                rc = GLD_FAILURE;
		goto done;
	}

	rc = usbnet_ctrl_write(usbnetp,
                                USBNET_CMD_SET_HW_MII,
                                0, 0, buf8, 0);
	if (rc < 0) {
                rc = GLD_FAILURE;
		goto done;
	}

        rc = GLD_SUCCESS;

done:
	mutex_exit(&usbnetp->usbnet_xmitlock);
	mutex_exit(&usbnetp->usbnet_intrlock);

	return (rc);
}

static  int
usbnet_start(gld_mac_info_t *usbnet_mac_info)
{
	int                     rc;
	uint8_t                 buf8[16];
        uint16_t                *buf16 = (uint16_t *) buf8;
	usb_pipe_policy_t       policy;
        usb_intr_req_t          *intr_request;
        usb_bulk_req_t          *bulk_request;
        struct usbnet           *usbnetp;

        usbnetp = (struct usbnet *) usbnet_mac_info->gldm_private;

	mutex_enter(&usbnetp->usbnet_intrlock);
	mutex_enter(&usbnetp->usbnet_xmitlock);

	usbnetp->usbnet_flags &= ~USBNET_RUNNING;
        usbnetp->usb_link_state = USBNET_PHY_STATE_INITIALIZING;

	rc = usbnet_ctrl_write(usbnetp,
                                USBNET_CMD_SET_SW_MII,
                                0, 0, buf8, 0);
	if (rc < 0) {
                rc = GLD_FAILURE;
		goto done;
	}

	buf16[0] = MII_CONTROL_RESET;
	rc = usbnet_ctrl_write(usbnetp,
                                USBNET_CMD_WRITE_MII_REG,
                                usbnetp->usb_physical_id[1],
                                MII_CONTROL,
                                buf16, 2);
	if (rc < 0) {
                rc = GLD_FAILURE;
		goto done;
	}

	/* Advertise that we can do full-duplex pause */
	buf16[0] = MII_ABILITY_10BASE_T |
                        MII_ABILITY_10BASE_T_FD |
                        MII_ABILITY_100BASE_TX |
                        MII_ABILITY_100BASE_TX_FD |
                        USBNET_ADVERTISE_CSMA |
                        USBNET_ADVERTISE_RESERVED;
	rc = usbnet_ctrl_write(usbnetp,
                                USBNET_CMD_WRITE_MII_REG,
                                usbnetp->usb_physical_id[1],
                                MII_AN_ADVERT,
                                buf16, 2);
	if (rc < 0) {
                rc = GLD_FAILURE;
		goto done;
	}

	buf16[0] = MII_CONTROL_ANE |
                        MII_CONTROL_RSAN;
	rc = usbnet_ctrl_write(usbnetp,
                                USBNET_CMD_WRITE_MII_REG,
                                usbnetp->usb_physical_id[1],
                                MII_CONTROL,
                                buf16, 2);
	if (rc < 0) {
                rc = GLD_FAILURE;
		goto done;
	}

	rc = usbnet_ctrl_write(usbnetp,
                                USBNET_CMD_SET_HW_MII,
                                0, 0, buf8, 0);
	if (rc < 0) {
                rc = GLD_FAILURE;
		goto done;
	}

	rc = usbnet_ctrl_write(usbnetp,
                                USBNET_CMD_WRITE_RX_CTL,
                                0x80, 0, buf8, 0);
	if (rc < 0) {
                rc = GLD_FAILURE;
		goto done;
	}

	/*
	 * USB Pipes Initialization
	 */

        usbnet_close_data_pipes(usbnetp);

	/*
	 * USB INTR Pipe Initialization
	 */

	policy.pp_max_async_reqs = USBNET_MAX_ASYNCH_REQS;

	rc = usb_pipe_open(usbnetp->usbnet_dip,
                                &usbnetp->usb_interrupt_descr,
                                &policy,
                                USB_FLAGS_SLEEP,
                                &usbnetp->usb_interrupt_pipe);

        if (rc != USB_SUCCESS) {
                usbneterror("usbnet_start: usb_pipe_open(INTR) failed: %d\n", rc);
                rc = GLD_FAILURE;
                goto done;
        }

        intr_request = usb_alloc_intr_req(usbnetp->usbnet_dip,
                                                0, USB_FLAGS_SLEEP);
        if (intr_request == NULL) {
                usbneterror("usbnet_start: usb_alloc_intr_request failed\n");
                rc = GLD_FAILURE;
                goto done;
        }

        intr_request->intr_len =
                        usbnetp->usb_interrupt_descr.wMaxPacketSize;
        intr_request->intr_client_private = (usb_opaque_t) usbnetp;
        intr_request->intr_attributes = USB_ATTRS_AUTOCLEARING |
                                                USB_ATTRS_SHORT_XFER_OK |
                                                USB_ATTRS_ONE_XFER;
        intr_request->intr_cb = usbnet_interrupt_pipe_callback;
        intr_request->intr_exc_cb = usbnet_interrupt_pipe_exception_callback;

        rc = usb_pipe_intr_xfer(usbnetp->usb_interrupt_pipe,
                                        intr_request,
                                        0);

        if (rc != USB_SUCCESS) {
                usbneterror("usbnet_start: usb_pipe_intr_xfer failed: %d\n", rc);
                usb_free_intr_req(intr_request);
                rc = GLD_FAILURE;
                goto done;
        }

        usbnetp->usbnet_intr_active = 1;

	/*
	 * USB BULK OUT Pipe Initialization
	 */

	policy.pp_max_async_reqs = USBNET_MAX_ASYNCH_REQS;

	rc = usb_pipe_open(usbnetp->usbnet_dip,
                                &usbnetp->usb_bulk_out_descr,
                                &policy,
                                USB_FLAGS_SLEEP,
                                &usbnetp->usb_bulk_out_pipe);

        if (rc != USB_SUCCESS) {
                usbneterror("usbnet_start: usb_pipe_open(BULK OUT) failed: %d\n", rc);
                rc = GLD_FAILURE;
                goto done;
        }

        usbnetp->usbnet_send_firstptr = NULL;
        usbnetp->usbnet_send_lastptr = NULL;
        usbnetp->usbnet_send_inprogress = 0;
        usbnetp->usbnet_send_active = 1;

	/*
	 * USB BULK IN Pipe Initialization
	 */

	policy.pp_max_async_reqs = USBNET_MAX_ASYNCH_REQS;

	rc = usb_pipe_open(usbnetp->usbnet_dip,
                                &usbnetp->usb_bulk_in_descr,
                                &policy,
                                USB_FLAGS_SLEEP,
                                &usbnetp->usb_bulk_in_pipe);

        if (rc != USB_SUCCESS) {
                usbneterror("usbnet_start: usb_pipe_open(BULK IN) failed: %d\n", rc);
                rc = GLD_FAILURE;
                goto done;
        }

        bulk_request = usb_alloc_bulk_req(usbnetp->usbnet_dip,
                                                ETHERMAX,
                                                USB_FLAGS_SLEEP);

        if (bulk_request == NULL) {
                usbneterror("usbnet_start: usb_alloc_bulk_request failed\n");
                rc = GLD_FAILURE;
                goto done;
        }

        bulk_request->bulk_len = ETHERMAX;
        bulk_request->bulk_timeout = USBNET_DEFAULT_TIMEOUT;
        bulk_request->bulk_client_private = (usb_opaque_t) usbnetp;
        bulk_request->bulk_attributes = USB_ATTRS_AUTOCLEARING |
                                                USB_ATTRS_SHORT_XFER_OK;
        bulk_request->bulk_cb = usbnet_bulk_in_pipe_callback;
        bulk_request->bulk_exc_cb = usbnet_bulk_in_pipe_exception_callback;

        rc = usb_pipe_bulk_xfer(usbnetp->usb_bulk_in_pipe,
                                bulk_request,
                                0);

        if (rc != USB_SUCCESS) {
                usbneterror("usbnet_start: usb_pipe_bulk_xfer failed: %d\n", rc);
                usb_free_bulk_req(bulk_request);
                rc = GLD_FAILURE;
                goto done;
        }

        usbnetp->usbnet_recv_active = 1;

	/*
	 * Init done.
	 */

	usbnetp->usbnet_flags |= USBNET_RUNNING;

        rc = GLD_SUCCESS;

done:
	mutex_exit(&usbnetp->usbnet_xmitlock);
	mutex_exit(&usbnetp->usbnet_intrlock);

	return rc;
}

static  int
usbnet_stop(gld_mac_info_t *usbnet_mac_info)
{
        int             rc;
	uint8_t         buf8[16];
        struct usbnet   *usbnetp;

        usbnetp = (struct usbnet *) usbnet_mac_info->gldm_private;

	mutex_enter(&usbnetp->usbnet_intrlock);
	mutex_enter(&usbnetp->usbnet_xmitlock);

	rc = usbnet_ctrl_write(usbnetp,
                                USBNET_CMD_WRITE_RX_CTL,
                                0x80, 0, buf8, 0);
	if (rc < 0) {
                rc = GLD_FAILURE;
		goto done;
	}

	usbnetp->usbnet_flags &= ~USBNET_RUNNING;

        usbnet_close_data_pipes(usbnetp);

        rc = GLD_SUCCESS;

done:
	mutex_exit(&usbnetp->usbnet_xmitlock);
	mutex_exit(&usbnetp->usbnet_intrlock);

	return (rc);
}


static  int
usbnet_get_stats(gld_mac_info_t *usbnet_mac_info, struct gld_stats *usbnet_stats)
{
	return (GLD_SUCCESS);
}

static  int
usbnet_send(gld_mac_info_t *usbnet_mac_info, mblk_t *mblkptr)
{
        int             rc;
        uint_t          len;
        struct usbnet   *usbnetp;

        usbnetp = (struct usbnet *) usbnet_mac_info->gldm_private;

        if (usbnetdebug > 0)
                usbneterror("usbnet_send STARTED. Usbnetp: 0x%08p Mblk: 0x%08p\n",
                                (void *) usbnetp,
                                (void *) mblkptr);

	mutex_enter(&usbnetp->usbnet_xmitlock);

        rc = usbnet_glue_stream_msg(mblkptr,
                                        ETHERMAX,
                                        usbnetp->usbnet_send_buf,
                                        &len);

        if (mblkptr != NULL)
                freemsg(mblkptr);

        if (rc == DDI_SUCCESS && len > 0 &&
                (usbnetp->usbnet_flags & USBNET_RUNNING) != 0 &&
                usbnetp->usbnet_send_active != 0) {

                int             do_send;

                if (usbnetdebug > 0)
                        usbneterror("usbnet_send SENDING DATA. Usbnetp: 0x%08p Mblk: 0x%08p Len: %d\n",
                                        (void *) usbnetp,
                                        (void *) mblkptr,
                                        len);

                if (usbnetdebug > 2)
                        usbnet_dump_data(usbnetp->usbnet_send_buf,
                                                len);

                mblkptr = allocb(len, 0);

                if (mblkptr == NULL) {
                        mutex_exit(&usbnetp->usbnet_xmitlock);
                        usbneterror("usbnet_send: allocb failed: %d\n",
                                        len);
                        return (GLD_FAILURE);
                }

                mblkptr->b_next = NULL;

                bcopy((uint8_t *) usbnetp->usbnet_send_buf,
                        (uint8_t *) mblkptr->b_wptr,
                        len);

                mblkptr->b_wptr += len;

                /* If there is no active send operation, start one,
                        else stick the message on the send queue */
                if (usbnetp->usbnet_send_inprogress == 0) {

                        usbnetp->usbnet_send_inprogress = 1;
                        do_send = 1;

                } else {

                        if (usbnetp->usbnet_send_firstptr == NULL) {
                                usbnetp->usbnet_send_firstptr = mblkptr;
                                usbnetp->usbnet_send_lastptr = mblkptr;
                        } else {
                                usbnetp->usbnet_send_lastptr->b_next = mblkptr;
                                usbnetp->usbnet_send_lastptr = mblkptr;
                        }

                        do_send = 0;

                }

                mutex_exit(&usbnetp->usbnet_xmitlock);

                if (do_send != 0) {

                        usb_bulk_req_t          *bulk_request;

                        bulk_request = usb_alloc_bulk_req(usbnetp->usbnet_dip,
                                                                0, USB_FLAGS_SLEEP);

                        if (bulk_request == NULL) {
                                usbneterror("usbnet_send: usb_alloc_bulk_request failed\n");
                                mutex_enter(&usbnetp->usbnet_xmitlock);
                                usbnetp->usbnet_send_inprogress = 0;
                                mutex_exit(&usbnetp->usbnet_xmitlock);
                                return (GLD_FAILURE);
                        }

                        bulk_request->bulk_len = len;
                        bulk_request->bulk_data = mblkptr;
                        bulk_request->bulk_timeout = USBNET_DEFAULT_TIMEOUT;
                        bulk_request->bulk_client_private = (usb_opaque_t) usbnetp;
                        bulk_request->bulk_attributes = USB_ATTRS_AUTOCLEARING;
                        bulk_request->bulk_cb = usbnet_bulk_out_pipe_callback;
                        bulk_request->bulk_exc_cb = usbnet_bulk_out_pipe_exception_callback;

                        rc = usb_pipe_bulk_xfer(usbnetp->usb_bulk_out_pipe,
                                                        bulk_request,
                                                        0);

                        if (rc != USB_SUCCESS) {
                                usbneterror("usbnet_send: usb_pipe_bulk_xfer failed: %d\n", rc);
                                mutex_enter(&usbnetp->usbnet_xmitlock);
                                usbnetp->usbnet_send_inprogress = 0;
                                mutex_exit(&usbnetp->usbnet_xmitlock);
                                usb_free_bulk_req(bulk_request);
                                return (GLD_FAILURE);
                        }

                }

        } else {
                mutex_exit(&usbnetp->usbnet_xmitlock);
                return (GLD_FAILURE);
        }

	return (GLD_SUCCESS);
}

static  int
usbnet_set_multicast(gld_mac_info_t *usbnet_mac_info, uint8_t *multicastaddr, int multicastflag)
{
        int             rc;
	uint8_t         buf8[16];
        struct usbnet   *usbnetp;

        usbnetp = (struct usbnet *) usbnet_mac_info->gldm_private;

	mutex_enter(&usbnetp->usbnet_intrlock);
	mutex_enter(&usbnetp->usbnet_xmitlock);

        /*
         * Fow now just initialize default broadcast address
         */
	rc = usbnet_ctrl_write(usbnetp,
                                USBNET_CMD_WRITE_RX_CTL,
                                0x8c, 0, buf8, 0);
	if (rc < 0) {
                rc = GLD_FAILURE;
		goto done;
	}

        rc = GLD_SUCCESS;

done:
	mutex_exit(&usbnetp->usbnet_xmitlock);
	mutex_exit(&usbnetp->usbnet_intrlock);

	return (rc);
}

static  int
usbnet_set_promiscuous(gld_mac_info_t *usbnet_mac_info, int promiscuousflag)
{
        int             rc;
	uint8_t         buf8[16];
        struct usbnet   *usbnetp;

        usbnetp = (struct usbnet *) usbnet_mac_info->gldm_private;

	mutex_enter(&usbnetp->usbnet_intrlock);
	mutex_enter(&usbnetp->usbnet_xmitlock);

        /*
         * Either start promoscuous mode or
         * initialize default broadcast address
         */
	rc = usbnet_ctrl_write(usbnetp,
                                USBNET_CMD_WRITE_RX_CTL,
                                (promiscuousflag ? 0x8d : 0x8c),
                                0, buf8, 0);
	if (rc < 0) {
                rc = GLD_FAILURE;
		goto done;
	}

        rc = GLD_SUCCESS;

done:
	mutex_exit(&usbnetp->usbnet_xmitlock);
	mutex_exit(&usbnetp->usbnet_intrlock);

	return (rc);
}
